from ..base import is_empty
from ..resource.bd_region import cities
from ..resource.bd_region import regions
from ..resource.patterns import AddressPattern
from ..util.district import District
from .decorator import check_null
from .util import get_shortest_element
from .util import re_fast


def get_single_list(start, length, step):
  """
  获取单个起点的位置端点列表，如果起始位置不是零，则会返回0到起始位置的列表
  Args:
    start: 起始位置
    length: 总长度
    step: 步长
  Examples:
    >>> get_single_list(0, 7, 3) == [[0, 3], [3, 6], [6, 7]]
    >>> get_single_list(1, 7, 3) == [[0, 1], [1, 4], [4, 7]]
  """
  if start > 0:
    res = [[0, start]]
  else:
    res = []
  for i in range(start, length, step):
    if i + step <= length:
      res.append([i, i + step])
    else:
      res.append([i, length])
  return res


def get_breaks(length, step):
  """
  通过长度和步长获所有可能的取端点列表的列表
  Examples:
    >>> get_breaks(10, 3) == [
    >>>   [[0, 3], [3, 6], [6, 9], [9, 10]],
    >>>   [[0, 1], [1, 4], [4, 7], [7, 10]],
    >>>   [[0, 2], [2, 5], [5, 8], [8, 10]]
    >>> ]
  """
  res = []
  for start in range(0, step):
    res.append(get_single_list(start, length, step))
  return res


def get_list_by_position(string: str, breaks: list):
  """
  按照位置信息将字符串拆解为多个字符串的列表
  Args:
    string: 需要拆分的字符串
    breaks: 位置集合列表的列表（左闭右开），[[1, 5], [5, 10], [10, 15]]
  """
  res = []
  for brk in breaks:
    res.append(string[brk[0]:brk[1]])
  return res


def drop_repeat_element(strs: list):
  """删除列表中连续重复的元素"""
  s = ''
  last = ''
  for i in strs:
    if i != last:
      s += i
      last = i
  return s


def drop_repeat_string_by_step(string, step):
  """根据指定的小块长度，获取去重后的字符串"""
  _list = []
  length = len(string)
  for breaks in get_breaks(length, step):
    res = get_list_by_position(string, breaks)
    str_drop = drop_repeat_element(res)
    _list.append(str_drop)

  return get_shortest_element(_list)


@check_null()
def drop_repeat_string(string,
                       min_length=3,
                       max_length=None):
  """
  删除连续重复的字符串，按照step从大到小删除重复字符，返回去重后最短的字符串

  Args:
    string: 要处理的字符串
    min_length: 识别的最短长度，默认为3
    max_length: 识别的最长长度，默认不限制

  Examples:
    >>> drop_repeat_string('上海市上海市杨浦区')
    '上海市杨浦区'
  """
  if len(string) < min_length * 2:
    return string
  _list = []
  if not max_length:
    max_length = len(string) // 2
  for step in range(max_length, min_length - 1, -1):
    str_drop = drop_repeat_string_by_step(string, step)
    _list.append(str_drop)
    string = get_shortest_element(_list)
  return string


@check_null(default_rv=[])
def extract_possible_region(string: str) -> list:
  """从字符串中提取可能的城市、区县"""
  if not isinstance(string, str):
    return []
  ds = District()
  ls = []
  city_num, region_num = 0, 0
  # 根据正则表达式提取城市和区县
  for pattern in [*AddressPattern.cities, *AddressPattern.regions]:
    if res := re_fast(pattern, string, warning=False):
      if ds.is_city(res):
        ls.append(res)
        city_num += 1
      if ds.is_region(res):
        ls.append(res)
        region_num += 1
  # 没提取出区县时，按照字符串匹配尝试提取
  if region_num == 0:
    for cr in regions():
      if cr in string:
        ls.append(cr)
        break
  # 没提取出城市时，按照字符串匹配尝试提取
  if city_num == 0:
    for cr in cities():
      if cr in string:
        ls.append(cr)
        break
  return list(set(ls))


def get_city_and_region(string) -> tuple:
  """从字符串中提取城市、区县"""
  ds = District()
  city_list, region_list = [], []
  ls = extract_possible_region(string)
  for i in ls:
    if ds.is_city(i):
      city_list.append(i)
    if ds.is_region(i):
      region_list.append(i)
  if region_list and not city_list:
    for r in region_list:
      if _city := ds.city(r, warning=False):
        city_list.append(_city)
  return (
    city_list[0] if city_list else None,
    region_list[0] if region_list else None,
    city_list,
    region_list
  )


def extract_city(string: str, na=None):
  """从字符串中提取城市（可能包含县级市）"""
  res = get_city_and_region(string)
  rv = res[0] or res[1]
  if is_empty(rv):
    return na
  return rv



import csv
import os

import pandas as pd
from shapely.geometry.base import BaseGeometry

from ..geometry.df import auto2shapely
from ..geometry.util import wkb_dumps
from ..util.os import ensure_dirpath_exist
from ..util.os import extension
from ..util.util import first_notnull_value
from . import ALL_EXTS
from .transformer import df_iter


def to_csv_by_line(data: list, filename: str):
  """
  逐行写入csv文件

  Args:
    data: 要写入的数据列表
    filename: 文件名
  """
  with open(filename, 'a') as f:
    csv_write = csv.writer(f, dialect='excel')
    csv_write.writerow(data)


def to_sheets(data: dict, filename: str, index=False):
  """
  将多个dataframe保存到不同的sheet中

  Args:
    data: 要保存的数据集，格式为：{sheet_name: DataFrame}
    filename: 要保存的文件名
    index: 是否保存index
  """
  assert isinstance(data, dict), 'data must be dict'
  with pd.ExcelWriter(filename) as writer:
    for sheet_name, _df in data.items():
      _df.to_excel(writer, sheet_name, index=index)


def to_parquet(df, filepath, index):
  """保存parquet文件，默认将shapely对象转换为wkb格式"""
  shapely_columns = [
    c for c in df if isinstance(first_notnull_value(df[c]), BaseGeometry)
  ]
  for c in shapely_columns:
    df[c] = df[c].apply(wkb_dumps)
  df.to_parquet(filepath, index=index)


def to_file(df: pd.DataFrame, filepath,
            *,
            index=False,
            log=True,
            encoding=None):
  """
  根据文件扩展名，将Dataframe保存为文件

  Args:
    df: 要保存的Dataframe
    filepath: 文件路径，包含扩展名
    index: 是否保存索引，默认不保存
    log: 是否打印保存信息
    encoding: 保存文件的编码
  """
  ensure_dirpath_exist(filepath)
  if log:
    print(f'Saving: {filepath}, Rows：{df.shape[0]}')
  df = df.copy()
  ex = extension(filepath)
  assert ex in ALL_EXTS, f'不支持的文件扩展名：{ex}'

  if ex == '.csv':
    df.to_csv(filepath, index=index, encoding=encoding)
  if ex in ('.pa', '.parquet'):
    to_parquet(df, filepath, index=index)
  if ex in ('.xlsx', '.xls'):
    df.to_excel(filepath, index=index)
  if ex == 'json':
    df.to_json(filepath, orient='records')
  if ex in ('.shp', '.geojson'):
    df = auto2shapely(df)
    df.to_file(filepath, encoding=encoding,
               driver='GeoJSON' if ex == '.geojson' else None)
  if ex in ('.sav', '.zsav'):
    import pyreadstat
    pyreadstat.write_sav(df, filepath)
  if ex == '.feather':
    df.to_feather(filepath, index=index)
  if ex == '.pickle':
    df.to_pickle(filepath)


def to_parts_file(df, dirpath,
                  chunksize=None,
                  parts=None,
                  to_ext='.csv',
                  **kwargs):
  """
  将Dataframe保存为多个文件

  Args:
    df: 要拆分保存的DataFrame
    dirpath: 文件保存的目录
    chunksize: 拆分保存的文件大小
    parts: 拆分保存的文件数量
    to_ext: 文件扩展名
  """
  for i, _df in enumerate(df_iter(df, chunksize=chunksize, parts=parts)):
    savefile = os.path.join(dirpath, f'part_{str(i).zfill(6)}{to_ext}')
    to_file(_df, savefile, **kwargs)



import warnings

import pandas as pd

from ..base import is_empty


def get_graph_dict(
    graph_df: pd.DataFrame,
    c_key='id',
    c_level_type='level_type',
    c_parent_key='parent_id',
    c_parent_type='parent_type'
) -> dict:
  """
  将图结构的Dataframe转为字典格式，提高查询效率

  Args:
    graph_df: 图结构的数据
    c_key: 索引列
    c_level_type: 层级分类列
    c_parent_key: 对应的父级索引列
    c_parent_type: 对应的父级类型列
  """
  graph_df = graph_df[[c_key, c_level_type, c_parent_key]]
  df_temp = graph_df[[c_key, c_level_type]]
  df_temp.columns = [c_parent_key, c_parent_type]
  graph_df = graph_df.merge(df_temp)
  # 构造查询字典
  return graph_df.set_index(c_key).to_dict()


def query_from_graph(
    key,
    graph_data: (pd.DataFrame, dict),
    c_key='id',
    c_level_type='level_type',
    c_parent_key='parent_id',
    max_depth=10,
) -> dict:
  """
  从图数据中查询全部当前节点的全部父级节点

  Args:
    key: 要查询的节点
    graph_data: 要查询的数据集
    c_key: 子节点关键列的字段名
    c_level_type: 等级类型字段名
    c_parent_key: 父节点字段名
    max_depth: 最大深度
  """

  # 构造查询字典
  c_parent_type = 'parent_type'
  if isinstance(graph_data, pd.DataFrame):
    warnings.warn(
        'graph_data 是一个 DataFrame, '
        '请使用 ricco.etl.graph.get_graph_dict 函数构造查询字典后传入，以提高查询效率')
    graph_dict = get_graph_dict(
        graph_df=graph_data, c_key=c_key, c_level_type=c_level_type,
        c_parent_key=c_parent_key, c_parent_type=c_parent_type)
  elif isinstance(graph_data, dict):
    graph_dict = graph_data
  else:
    raise TypeError('graph_data 类型错误，请传入 DataFrame 或 dict')
  level_type = graph_dict[c_level_type].get(key)
  if not level_type:
    return {}
  res = {level_type: key}
  # 循环查询上级关系
  depth = 0
  while depth < max_depth:
    parent_key = graph_dict[c_parent_key].get(key)
    parent_type = graph_dict[c_parent_type].get(key)
    if is_empty(parent_key):
      break
    res[parent_type] = parent_key
    key = parent_key
    depth += 1
  return res



from functools import lru_cache

import pandas as pd

from ..base import not_empty
from ..resource import P_BD_REGION
from ..util.util import rstrip_d0

EX = ['新城区']


@lru_cache()
def get_bd_region(start_level='区县') -> pd.DataFrame:
  """获取bd_region表并转为宽表；同时保存为全局变量，避免多次IO"""
  default_level = ['街镇', '区县', '城市', '省份']
  levels = default_level[default_level.index(start_level):]

  df = pd.read_csv(P_BD_REGION)
  for c in ['parent_id', 'id']:
    df[c] = df[c].astype(str)
    df[c] = df[c].apply(rstrip_d0)
  df = df[df['parent_id'].notna()]
  df = df[df['level_type'].isin([1, 2, 3, 4])]
  df['level_type'] = df['level_type'].replace(
      to_replace={1: '省份', 2: '城市', 3: '区县', 4: '街镇'}
  )

  df_prov = df[df['level_type'] == '省份']
  df_prov = df_prov[['id', 'name', 'short_name']]
  df_prov.columns = ['省份id', '省份名称', '省份简称']
  dfs = {
    '省份': df_prov.copy()
  }

  for i, l in enumerate(levels[:-1]):
    _df = df[df['level_type'] == l]
    _df = _df[['id', 'name', 'short_name', 'parent_id']]
    _df.columns = [f'{l}id', f'{l}名称', f'{l}简称', f'{levels[i + 1]}id']
    dfs[l] = _df.copy()

  df_res = dfs[levels[0]]
  for i in range(len(levels) - 1):
    df_res = df_res.merge(
        dfs[levels[i + 1]],
        on=f'{levels[i + 1]}id',
        how='outer',
    )
  return df_res


@lru_cache()
def cities():
  """城市列表，按照长度排序"""
  ls = get_bd_region()['城市名称'].unique().tolist()
  ls = [i for i in ls if not_empty(i)]
  return sorted(ls, key=lambda x: len(x), reverse=True)


@lru_cache()
def regions():
  """区县列表，按照长度排序"""
  ls = get_bd_region()['区县名称'].unique().tolist()
  ls = [i for i in ls if not_empty(i) and i not in EX]
  return sorted(ls, key=lambda x: len(x), reverse=True)


@lru_cache()
def city_region_list():
  """城市和区县列表，按照长度排序"""
  df = get_bd_region()
  ls = [*df['城市名称'].unique().tolist(), *df['区县名称'].unique().tolist()]
  ls = [i for i in ls if not_empty(i) and i not in EX]
  return sorted(ls, key=lambda x: len(x), reverse=True)



import logging
import os
import shutil
import warnings
import zipfile
from datetime import datetime
from datetime import timedelta
from pathlib import Path

from ..base import ensure_list


def protect_dir(path):
  if os.path.isdir(path):
    if path == '/':
      raise ValueError('不可指定根目录')
    dirs = [
      '/bin', '/sbin', '/usr', '/etc', '/dev', '/Applications', '/Library',
      '/Network', '/System', '/Volumes', '/cores', '/private'
    ]
    for pre in dirs:
      if os.path.abspath(path).startswith(fr'{pre}'):
        raise ValueError('系统目录下的文件不可删除')


def ext(filepath):
  """获取文件扩展名"""
  return extension(filepath)


def extension(filepath):
  """获取文件扩展名"""
  return os.path.splitext(filepath)[1]


def fn(filepath):
  """路径及文件名（不含扩展名）"""
  return path_name(filepath)


def path_name(filepath):
  """路径及文件名（不含扩展名）"""
  return os.path.splitext(filepath)[0]


def split_path(filepath, abspath=False):
  """将文件路径拆分为文件夹路径、文件名、扩展名三部分"""
  if abspath:
    filepath = os.path.abspath(filepath)
  basename = os.path.basename(filepath)
  return os.path.dirname(filepath), path_name(basename), extension(basename)


def ensure_dir(dirpath: str):
  """确保路径为文件夹格式（以斜杠“/”结尾）"""
  if not dirpath.endswith('/'):
    return dirpath + '/'
  return dirpath


def remove_path(path, log=True):
  """删除文件或文件夹"""
  if os.path.exists(path):
    os.remove(path) if os.path.isfile(path) else shutil.rmtree(path)
    if log:
      logging.warning(f'Deleted:"{path}"')
  else:
    warnings.warn(f'Path not found:"{path}"')


def find_undefined_dirs(dirpath):
  """获取不存在的目录列表"""
  dirs = []
  dirpath = os.path.abspath(dirpath)
  while not os.path.exists(dirpath):
    dirs.append(dirpath)
    dirpath = os.path.dirname(dirpath)
  dirs.reverse()
  return dirs


def ensure_dirpath_exist(filepath):
  """确保目录存在，不存在则创建"""
  dir_path = os.path.dirname(filepath)
  if os.path.exists(dir_path):
    return
  undefined_dirs = find_undefined_dirs(dir_path)
  for p in undefined_dirs:
    os.makedirs(p)
    logging.warning(f'Created: "{p}"')


def file2zip(filepath, overwrite=False, delete_origin=False):
  """将一个文件压缩为zip格式"""
  zfn = path_name(filepath) + '.zip'
  if os.path.exists(zfn) and not overwrite:
    raise FileExistsError(f'"{zfn}"已存在')
  logging.warning(f'Compressing: "{filepath}"')
  with zipfile.ZipFile(zfn, 'w', zipfile.ZIP_DEFLATED) as zipf:
    zipf.write(filepath, arcname=os.path.basename(filepath))
  if delete_origin:
    remove_path(filepath)


def dir2zip(dir_path, overwrite=False, delete_origin=False):
  """
  压缩文件夹

  Args:
    dir_path: 文件夹路径
    overwrite: 是否覆盖已有文件
    delete_origin: 是否删除原文件
  """
  protect_dir(dir_path)
  zfn = f'{dir_path}.zip'
  if os.path.exists(zfn) and not overwrite:
    raise FileExistsError(f'"{zfn}"已存在')
  z = zipfile.ZipFile(zfn, 'w', zipfile.ZIP_DEFLATED)
  for dirpath, _, filenames in os.walk(dir_path):
    for filename in filenames:
      f_o = str(os.path.join(dirpath, filename))
      f_i = str(os.path.join(os.path.split(dirpath)[-1], filename))
      z.write(f_o, f_i)
  z.close()
  if delete_origin:
    remove_path(dir_path)


def count_files(dir_path):
  """统计文件夹中文件的数量"""
  num = 0
  for dirpath, dirnames, filenames in os.walk(dir_path):
    for filename in filenames:
      if filename != '.DS_Store':
        num += 1
  return num


def rm_scratch_file(dir_path, days, recursive=False, rm_hidden_file=False,
                    exts=None):
  """删除指定天数前修改过的文件"""
  protect_dir(dir_path)
  timeline = datetime.now() - timedelta(days)
  # 删除过期文件
  for file in dir_iter(dir_path, abspath=True, recursive=recursive,
                       ignore_hidden_files=not rm_hidden_file, exts=exts):
    m_time = os.path.getmtime(file)
    m_time = datetime.fromtimestamp(m_time)
    if m_time <= timeline:
      os.remove(file)
      logging.warning(f'Deleted:"{file}"')
  # 删除空文件夹
  for dirpath, dirnames, _ in os.walk(dir_path):
    for dirname in dirnames:
      dir_full_path = os.path.join(dirpath, dirname)
      num = count_files(dir_full_path)
      if num == 0:
        remove_path(dir_full_path)


def remove_ds_store(dir_path):
  """删除文件夹中的.DS_Store文件"""
  protect_dir(dir_path)
  for i in dir_iter(dir_path, recursive=True, ignore_hidden_files=False):
    if i.endswith('.DS_Store'):
      remove_path(i)


def dir_iter(root,
             exts: (list, str) = None,
             abspath=False,
             recursive=False,
             ignore_hidden_files=True):
  """
  文件夹中的文件路径生成器，用于遍历文件夹中的文件

  Args:
    root: 文件目录
    exts: 文件扩展名，不指定则返回所有文件
    abspath: 是否返回绝对路径
    recursive: 是否循环遍历更深层级的文件，默认只返回当前目录下的文件
    ignore_hidden_files: 是否忽略隐藏文件
  """
  if exts:
    exts = ensure_list(exts)

  for dirpath, _, filenames in os.walk(root):
    for _name in filenames:
      if ignore_hidden_files and _name.startswith('.'):
        continue
      filepath = os.path.join(dirpath, _name)
      filepath = os.path.abspath(filepath) if abspath else filepath
      # 不符合扩展名要求忽略
      if exts and extension(_name) not in exts:
        continue
      # 仅需要当前目录时，dirpath和root不相同的过滤掉
      if not recursive and dirpath != root:
        continue
      yield filepath


def dir_iter_list(root,
                  exts: (list, str) = None,
                  abspath=False,
                  recursive=False,
                  reverse=False):
  """
  文件夹中的文件路径列表

  Args:
    root: 文件目录
    exts: 文件扩展名，不指定则返回所有文件
    abspath: 是否返回绝对路径
    recursive: 是否循环遍历更深层级的文件，默认只返回当前目录下的文件
    reverse: 路径列表是否倒序
  """
  path_list = []
  for p in dir_iter(root, exts, abspath, recursive, ignore_hidden_files=True):
    path_list.append(p)
  return sorted(path_list, reverse=reverse)


def single_ext(path_list):
  ext_list = list(set([extension(p) for p in path_list]))
  if len(ext_list) == 1:
    return ext_list[0]


class OssPath:
  def __init__(self, path, bucket=None):
    """
    OSS路径类，确保不同类型的地址属性统一

    Args:
      path: 路径，完整路径或bucket后的部分路径
      bucket: 默认从完整路径中获取，当为部分路径时需指定Bucket
    """

    if path.startswith('oss://'):
      pass
    elif bucket:
      path = os.path.join('oss:///', bucket, path)
    else:
      raise Exception('Bucket name is required')

    self.parts = Path(path).parts

    if bucket:
      assert bucket == self.parts[1], 'Bucket name does not match'

  @property
  def bucket(self):
    """Bucket"""
    return self.parts[1]

  @property
  def driver(self):
    """Driver"""
    return self.parts[0]

  @property
  def path_short(self):
    """短路径"""
    if len(self.parts) > 2:
      return str(os.path.join(*self.parts[2:]))
    return ''

  @property
  def path_full(self):
    """完整路径"""
    return str(os.path.join(f'oss:///{self.bucket}', self.path_short))



"""地理/geometry相关"""

from .coord_trans import coord_trans_geom
from .coord_trans import coord_trans_x2y
from .coord_trans import coord_transformer
from .df import auto2shapely
from .df import buffer
from .df import geojson2shapely
from .df import get_area
from .df import lnglat2shapely
from .df import lnglat2wkb
from .df import lnglat2wkt
from .df import mark_tags_v2
from .df import nearest_kdtree
from .df import nearest_neighbor
from .df import projection
from .df import shapely2geojson
from .df import shapely2lnglat
from .df import shapely2wkb
from .df import shapely2wkt
from .df import shapely2x
from .df import spatial_agg
from .df import wkb2lnglat
from .df import wkb2shapely
from .df import wkb2wkt
from .df import wkt2lnglat
from .df import wkt2shapely
from .df import wkt2wkb
from .util import _projection_lnglat
from .util import distance
from .util import epsg_from_lnglat
from .util import geojson_dumps
from .util import geojson_loads
from .util import get_epsg
from .util import get_inner_point
from .util import infer_geom_format
from .util import is_geojson
from .util import is_shapely
from .util import is_wkb
from .util import is_wkt
from .util import wkb_dumps
from .util import wkb_loads
from .util import wkt_dumps
from .util import wkt_loads



import math
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta

from ..base import ensure_list
from ..base import is_empty
from ..util.assertion import assert_not_null
from ..util.assertion import assert_series_unique
from ..util.decorator import check_null
from ..util.decorator import process_multi
from ..util.decorator import progress
from ..util.decorator import run_once
from ..util.decorator import timer
from ..util.util import and_
from ..util.util import fuzz_match
from ..util.util import or_
from ..util.util import to_float
from ..util.util import to_str_list
from .graph import get_graph_dict
from .graph import query_from_graph


def keep_best_unique(df: pd.DataFrame,
                     subset: (list, str),
                     value_cols: (str, list) = None
                     ) -> pd.DataFrame:
  """
  优化的去重函数，为保证数据的完整性，去重时优先去除指定列中的空值

  Args:
    df: 要去重的Dataframe
    subset: 按照哪些列去重
    value_cols: 优先去除那些列的空值，该列表是有顺序的
  """
  subset = ensure_list(subset)

  if not value_cols:
    value_cols = [i for i in df if i not in subset]
  value_cols = ensure_list(value_cols)

  return df.sort_values(
      value_cols, na_position='first'
  ).drop_duplicates(
      subset, keep='last'
  ).sort_index()


def table2dict(df: pd.DataFrame,
               key_col: str = None,
               value_col: (str, list) = None,
               orient: str = 'dict') -> dict:
  """
  DataFrame转字典

  Args:
    df:
    key_col: 生成key的列
    value_col: 生成value的列
    orient: 生成dict的方式，默认 'dict',还有 ‘list’, ‘series’, ‘split’, ‘records’, ‘index’
  """
  if not all([key_col, value_col]):
    columns = df.columns.tolist()
    key_col = columns[0]
    value_col = columns[1]
  df = df[df[key_col].notna()].set_index(key_col)
  if isinstance(value_col, list):
    df = df[value_col]
    return df.to_dict(orient=orient)
  else:
    df = df[[value_col]]
    return df.to_dict(orient=orient)[value_col]


def round_by_columns(df, columns: list):
  """对整列进行四舍五入，默认绝对值大于1的数值保留两位小数，小于1 的保留4位"""

  def _round(x):
    if is_empty(x):
      return np.nan
    if abs(x) >= 1:
      return round(x, 2)
    return round(x, 4)

  columns = ensure_list(columns)
  for c in columns:
    df[c] = df[c].apply(lambda x: _round(x))
  return df


def update_df(df: pd.DataFrame,
              new_df: pd.DataFrame,
              on: (str, list) = None,
              overwrite: bool = True,
              errors: str = 'ignore') -> pd.DataFrame:
  """
  根据某一列更新dataframe里的数据

  Args:
    df: 待升级的数据集
    new_df: 用于更新的DataFrame
    on: （可选参数）用于判断更新哪些行的列
    overwrite : （可选参数）控制如何处理原DataFrame在重叠位置上 **非空** 的值，默认为True

      * True: 默认值；使用 `other` DataFrame中的值覆盖原DataFrame中相应位置的值.
      * False: 只更新原DataFrame中重叠位置数据为 *空* 的值.
    errors: （可选参数）控制如何处理两个DataFrame同一位置都有值的行为，默认为 'ignore'

      * 'ignore': 默认值；DataFrame类型 df和other在同一个cell位置都是非NA值，
        使用other中的值替换df中的值。
      * 'raise': target和other都在同一位置包含非NA数据将抛出ValueError异常（'Data overlaps'）。
  """
  if on:
    df.set_index(on, inplace=True)
    df.update(new_df.set_index(on), overwrite=overwrite, errors=errors)
    df.reset_index(inplace=True)
  else:
    df.update(new_df, overwrite=overwrite, errors=errors)
  return df


def convert_date(df: pd.DataFrame,
                 columns: (str, list),
                 mode: str = 'first') -> pd.DataFrame:
  """
  将日期转为当月的第一天或最后一天

  Args:
    df: 要处理的DataFrame
    columns: 要转换的列
    mode: 'first' or 'last'
  """

  @check_null()
  def trans(x):
    return datetime(x.year, x.month, x.day)

  from pandas.tseries.offsets import MonthEnd
  assert mode in ('first', 'last'), "可选参数为first or last"
  columns = ensure_list(columns)
  for c in columns:
    df[c] = pd.to_datetime(df[c])
    if mode == 'first':
      df[c] = df[c].apply(lambda x: x.replace(day=1))
    else:
      df[c] = pd.to_datetime(df[c], format="%Y%m") + MonthEnd(1)
    df[c] = df[c].apply(trans)
  return df


@timer()
@process_multi
def fuzz_df(df: pd.DataFrame,
            col: str,
            target_series: (list, pd.Series),
            c_dst: str = None,
            valid_score=60) -> pd.DataFrame:
  """
  模糊匹配。为DataFrame中的某一列从某个集合中模糊匹配匹配相似度最高的元素

  Args:
    df: 输入的dataframe
    col: 要匹配的列
    target_series: 从何处匹配， list/pd.Series
    c_dst: 关联后输出的列名，默认为原列名+"_target"后缀
    valid_score: 相似度大于该值的才返回，默认60
  """
  if c_dst:
    assert isinstance(c_dst, str), 'c_dst must be str'
  target_series = to_str_list(target_series)
  _df = df[df[col].notna()][[col]].drop_duplicates(ignore_index=True)

  c_dst = c_dst or f'{col}_target'
  _df[[
    c_dst, 'normal_score', 'partial_score', 'weight_score'
  ]] = _df.parallel_apply(
      lambda r: fuzz_match(r[col], target_series, valid_score=valid_score),
      result_type='expand',
      axis=1)
  return df.merge(_df, on=col, how='left')


def convert_to_float(df: pd.DataFrame,
                     columns: (list, str),
                     rex_method: str = 'mean') -> pd.DataFrame:
  """
  提取字符串中的数值信息并转为float类型

  Args:
    df: 要转换的DataFrame
    columns: 要转换的列
    rex_method: 计算mean,max,min， 默认为mean
  """
  columns = ensure_list(columns)
  for c in columns:
    df[c] = df[c].apply(lambda x: to_float(x, rex_method=rex_method))
  return df


def _is_eq(series: pd.Series, value):
  """整合空和非空的判断条件"""
  return series.isna() if pd.isna(value) else (series == value)


def _compare_dtypes(df_main, df_other):
  """判断df_main和df_other的列类型是否一致，其中df_main的列是df_other列的子集"""
  for c in df_main:
    if df_main[c].dtypes != df_other[c].dtypes:
      warnings.warn(f'两个数据集"{c}"列类型不一致')


def filter_by_df(df: pd.DataFrame,
                 df_sizer: pd.DataFrame,
                 reverse: bool = False) -> pd.DataFrame:
  """从df中筛选出满足df_size的数据"""
  _compare_dtypes(df_sizer, df)
  df_sizer = df_sizer.drop_duplicates()
  cond = or_(*[
    and_(
        *[_is_eq(df[c], df_sizer[c][i]) for c in df_sizer]
    ) for i in df_sizer.index
  ])
  if reverse:
    return df[~cond]
  return df[cond]


def drop_by_df(df: pd.DataFrame, df_deleted: pd.DataFrame) -> pd.DataFrame:
  """根据一个df_deleted中的数据从df中删除"""

  _compare_dtypes(df_deleted, df)
  df_deleted = df_deleted.drop_duplicates()

  return df[~or_(*[
    and_(
        *[_is_eq(df[c], df_deleted[c][i]) for c in df_deleted]
    ) for i in df_deleted.index
  ])]


def wrap_dict(df: pd.DataFrame, c_srcs: list = None, c_dst: str = 'extra'):
  """将指定的多列合并成一个字典列
  Args:
    df: 要处理的数据集
    c_srcs: 需要包装成字典的多列, 如果不指定，则wrap所有的列
    c_dst: 处理后的数据存储的列名称，默认为 'extra'
  """
  df = df.copy()
  if not c_srcs:
    c_srcs = df.columns.tolist()
  extra = df[c_srcs]
  df = df.drop(c_srcs, axis=1)
  df[c_dst] = extra.to_dict('records')
  return df


@timer()
@progress
def expand_dict(df: pd.DataFrame, c_src: str):
  """展开字典为多列"""
  return pd.concat(
      [
        df.drop(c_src, axis=1),
        df[c_src].progress_apply(
            lambda x: pd.Series(x, dtype='object')
        ).set_index(df.index)
      ],
      axis=1
  )


def expand_dict_silent(df: pd.DataFrame, c_src: str):
  """展开字典为多列"""
  return pd.concat(
      [
        df.drop(c_src, axis=1),
        df[c_src].apply(
            lambda x: pd.Series(x, dtype='object')
        ).set_index(df.index)
      ],
      axis=1
  )


@timer()
def split_to_rows(df: pd.DataFrame, column: str, delimiter: str = '|'):
  """
  含有多值的列分拆成多行

  Args:
    df:
    column: 存在多值的列
    delimiter: 多个值之间的分割符
  """
  df = df.copy()
  df[column] = df[column].str.split(delimiter)
  return df.explode(column, ignore_index=True)


@timer()
def split_list_to_row(df: pd.DataFrame, column):
  """将列表列中列表的元素拆成多行"""
  return df.explode(column, ignore_index=True)


def dict2df(data: dict, c_key='key', c_value='value', as_index=False):
  """将字典转为dataframe"""
  df = pd.DataFrame()
  df[c_key] = data.keys()
  df[c_value] = data.values()
  return df.set_index(c_key) if as_index else df


@timer()
def is_changed(df_old: pd.DataFrame,
               df_new: pd.DataFrame,
               key_cols: (str, list) = None,
               value_cols: (str, list) = None,
               c_res: str = 'is_changed') -> pd.DataFrame:
  """
  判断新旧数据集中的每一条数据是否变化

  Args:
    df_old: 原始数据集
    df_new: 修改后的数据集
    key_cols: 关键列，不指定则以索引列为准
    value_cols: 要对比的列，默认出key_cols之外的其他列
    c_res: 对比结果列名，默认为“is_changed”
  """
  # 数据集及参数检验
  key_cols = ensure_list(key_cols)
  assert_series_unique(df_new, key_cols)
  if not value_cols:
    error_msg = 'Each datasets must have same columns.'
    assert set(df_new.columns) == set(df_old.columns), error_msg
    value_cols = [c for c in df_old if c not in key_cols]
  # 对比
  # 默认所有的都是更改的
  df_new = df_new.copy()
  df_new[c_res] = 'Changed'
  df_temp = df_old[[*key_cols, *value_cols]].merge(
      df_new[[*key_cols, *value_cols, c_res]],
      how='left',
      on=key_cols)
  # 合并后为空的为 NotFound
  df_temp[c_res] = df_temp[c_res].fillna('NotFound')

  # 对比每一个值，如果每一列的每一个值都相同，则认为是没有变化
  cond = and_(
      *[
        (df_temp[f'{c}_x'] == df_temp[f'{c}_y']) |
        (df_temp[f'{c}_x'].isna() & df_temp[f'{c}_y'].isna())
        for c in value_cols
      ]
  )
  df_temp.loc[cond, c_res] = 'NotChange'
  df_temp = df_temp[[*key_cols, c_res]]
  return df_old.merge(df_temp, how='left', on=key_cols)


def df_iter(df: pd.DataFrame, *, chunksize: int = None, parts: int = None):
  """
  DataFrame切片生成器

  Args:
    df: 要切片的DataFrame
    chunksize: 每次返回的大小，与parts必须指定一个
    parts: 返回的次数，与chunksize必须指定一个
  """
  assert any([chunksize, parts]), 'chunksize和parts必须指定一个'
  size = df.shape[0]
  if chunksize:
    parts = math.ceil(size / chunksize)
  else:
    chunksize = math.ceil(size / parts)
  for i in range(parts):
    low = i * chunksize
    high = (i + 1) * chunksize
    if i == parts - 1:
      yield df.iloc[low:, :]
    else:
      yield df.iloc[low: high, :]


def create_columns(df: pd.DataFrame, columns: list, value=None):
  """
  创建新的列，默认为空

  Args:
    df: DataFrame
    columns: 列名
    value: 值，默认为空
  """
  columns = ensure_list(columns)
  for c in columns:
    if c not in df:
      df[c] = value
  return df


@progress
def expand_graph(df: pd.DataFrame,
                 start_level,
                 c_key='id',
                 c_level_type='level_type',
                 c_parent_key='parent_id',
                 c_info=None):
  """
  将图数据展开为宽表

  Args:
    df: 要展开的图数据
    start_level: 开始展开的层级，一般是最低层级
    c_key: 关键列
    c_level_type: 层级类型列
    c_parent_key: 父级关键列
    c_info: 要保留的其他列
  """

  @run_once
  def asserts(_df, c_id, c_type, c_parent_id):
    assert isinstance(_df, pd.DataFrame)
    assert _df[c_id].is_unique
    assert_not_null(_df, c_type), f'{c_type}不能为空'
    assert _df[_df[c_id] == _df[c_parent_id]].empty, '父子不能相同'

  assert start_level in df[c_level_type].unique(), 'start_level不在level_type中'
  df_base = df.copy()

  # 构造查询字典
  graph_df = df.copy()
  asserts(graph_df, c_id=c_key, c_type=c_level_type,
          c_parent_id=c_parent_key)
  graph_df[c_level_type] = graph_df[c_level_type].apply(
      lambda lp: f'{lp}_{c_key}'
  )
  # 事先将Dataframe转为字典，提高查询性能
  graph_dict = get_graph_dict(graph_df=graph_df, c_key=c_key,
                              c_level_type=c_level_type,
                              c_parent_key=c_parent_key,
                              c_parent_type='parent_type')
  # 将索引列提取出来作为基础底表
  df = df[df[c_level_type] == start_level]
  df = df[[c_key]].reset_index(drop=True)
  # 将索引列逐个传入查询
  df['extra'] = df[c_key].progress_apply(
      lambda x: query_from_graph(
          key=x, graph_data=graph_dict, c_key=c_key,
          c_level_type=c_level_type, c_parent_key=c_parent_key)
  )
  # 展开最终结果
  df = expand_dict(df, 'extra')
  if c_info:
    c_info = ensure_list(c_info)
    for level_type in df_base[c_level_type].unique():
      df_level = df_base[[c_key, *c_info]]
      df_level.columns = [f'{level_type}_{c}' for c in df_level]
      if f'{level_type}_{c_key}' in df_level:
        df = df.merge(df_level, on=f'{level_type}_{c_key}', how='left')
  return df


def null_if_in(df, columns: (str, list), values: list):
  """
  如果列的值正好在values中定义的数组中，则将该列值设置为None

  Args:
    df:
    columns: 需要检查的列，一个或多少
    values: 所有需要标识为None的值
  """
  columns = ensure_list(columns)
  for c in columns:
    conditions = [df[c] == v for v in values]
    df.loc[or_(*conditions), c] = None
  return df


def fix_null(df: pd.DataFrame, columns: (str, list) = None):
  """将None/NaN/NaT等都替换为None"""
  if not columns:
    columns = df.columns
  columns = ensure_list(columns)
  for c in columns:
    df.loc[df[c].isna(), c] = None
  return df


def fillna_by_adding(df: pd.DataFrame, col: str):
  """以自增的方式填充空值"""
  df = df.copy()
  max_id = 0 if df[col].isna().all() else df[col].max()
  n = df[col].isna().sum()
  df.loc[df[col].isna(), col] = range(int(max_id + 1), int(max_id + 1 + n))
  df[col] = df[col].astype(int)
  assert_not_null(df, col)
  assert_series_unique(df, col)
  return df


def rolling_by_month(df, c_date, c_group, months: int, agg: dict):
  """
  按照日期进行滚动聚合，比如：分组对某个项目计算近6个月的成交面积

  Args:
    df: 输入的dataframe
    c_date: 日期列
    c_group: 分组列
    months: 窗口大小
    agg: 聚合函数，如：{'金额': 'sum'}
  """
  df = df.copy()
  df[c_date] = pd.to_datetime(df[c_date])
  # 起始日期为最小日期加上窗口时间
  date_start = df[c_date].min() + relativedelta(months=months)
  date_end = df[c_date].max()
  dfs = []
  while date_end >= date_start:
    date_l = date_start - relativedelta(months=months)
    _df = df[(df[c_date] >= date_l) & (df[c_date] < date_start)]
    _df = _df.groupby(c_group, as_index=False).agg(agg)
    _df[c_date] = date_start
    dfs.append(_df.copy())
    date_start += relativedelta(months=1)
  return pd.concat(dfs, ignore_index=True)



import logging
import time

# 64 位 id 的划分,通常机器位和数据位各为 5 位
WORKER_ID_BITS = 5  # 机器位
DATACENTER_ID_BITS = 5  # 数据位
SEQUENCE_BITS = 12  # 循环位

# 最大取值计算,计算机中负数表示为他的补码
MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS)  # 2**5 -1 =31
MAX_DATACENTER_ID = -1 ^ (-1 << DATACENTER_ID_BITS)

# 移位偏移计算
WORKER_ID_SHIFT = SEQUENCE_BITS
DATACENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS
TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATACENTER_ID_BITS

# X序号循环掩码
SEQUENCE_MASK = -1 ^ (-1 << SEQUENCE_BITS)

# Twitter 元年时间戳
TWEPOCH = 1288834974657

logger = logging.getLogger('雪花算法')


class IdGenerator:
  """
  用于生成IDS.

  Args:
    datacenter_id:数据id
    worker_id:机器id
    sequence:序列码
  """

  def __init__(self, datacenter_id=1, worker_id=1, sequence=0):
    assert 0 <= worker_id <= MAX_WORKER_ID, f'worker_id 值越界，范围为0-{MAX_WORKER_ID}'
    assert 0 <= datacenter_id <= MAX_DATACENTER_ID, f'datacenter_id 值越界，范围为0-{MAX_DATACENTER_ID}'

    self.worker_id = worker_id
    self.datacenter_id = datacenter_id
    self.sequence = sequence
    self.last_timestamp = -1  # 上次计算的时间戳
    time.sleep(0.001)  # 避免多次实例化调取方法导致id重复

  @staticmethod
  def _gen_timestamp():
    """生成整数时间戳"""
    return int(time.time() * 1000)

  def _til_next_millis(self, last_timestamp):
    """等到下一毫秒"""
    timestamp = self._gen_timestamp()
    while timestamp <= last_timestamp:
      timestamp = self._gen_timestamp()
    return timestamp

  def get_id(self):
    """获取新的ID"""
    # 获取当前时间戳
    timestamp = self._gen_timestamp()

    # 时钟回拨的情况
    if timestamp < self.last_timestamp:
      logging.error(
          'clock is moving backwards. Rejecting requests util {}'.format(
              self.last_timestamp))
      raise Exception('无效的系统时钟')

    if timestamp == self.last_timestamp:
      # 同一毫秒的处理。
      self.sequence = (self.sequence + 1) & SEQUENCE_MASK
      if self.sequence == 0:
        timestamp = self._til_next_millis(self.last_timestamp)
    else:
      self.sequence = 0

    self.last_timestamp = timestamp

    return (
        ((timestamp - TWEPOCH) << TIMESTAMP_LEFT_SHIFT) |
        (self.datacenter_id << DATACENTER_ID_SHIFT) |
        (self.worker_id << WORKER_ID_SHIFT)
    ) | self.sequence

  def get_id_str(self):
    return str(int(self.get_id()))



import re
import sys
import warnings
from ast import literal_eval
from itertools import groupby

import geojson
import numpy as np
import pandas as pd
from shapely import wkb
from shapely import wkt
from shapely.errors import GeometryTypeError
from shapely.errors import WKBReadingError
from shapely.geometry import GeometryCollection
from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPoint
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from shapely.geometry.base import BaseMultipartGeometry
from shapely.geometry.geo import ShapelyDeprecationWarning
from shapely.geos import WKTReadingError
from shapely.ops import unary_union
from shapely.validation import make_valid

from ..base import ensure_list
from ..base import is_empty
from ..base import not_empty
from ..util.decorator import check_null
from ..util.decorator import check_shapely
from ..util.decorator import check_str
from ..util.util import is_hex
from ..util.util import isinstance_in_list

warnings.simplefilter('ignore', category=ShapelyDeprecationWarning)

GEOM_FORMATS = ('wkb', 'wkt', 'shapely', 'geojson')


def crs_sh2000():
  """测绘院（上海2000）crs信息"""
  from ..resource.crs import CRS_SH2000
  return CRS_SH2000


def epsg_from_lnglat(lng, lat=0):
  """根据经纬度计算 UTM 区域 EPSG 代码"""
  lng = ensure_list(lng)
  lng = [i for i in lng if not_empty(i)]
  lng = np.median(lng)
  utm_band = str(int((np.floor((lng + 180) / 6) % 60) + 1)).zfill(2)
  return int(f'326{utm_band}' if lat >= 0 else f'327{utm_band}')


def lng_from_city(city: str):
  """获取城市所在的经度"""
  from ..resource.epsg_code import CITY_POINT
  assert len(city) >= 2, '城市名称过短'
  if city in CITY_POINT:
    return CITY_POINT[city]['lng']
  else:
    for _c in CITY_POINT.keys():
      if city in _c:
        return CITY_POINT[_c]['lng']
  warnings.warn(f'请补充"{city}"的epsg信息，默认返回经度113.0')
  return 113.0


def get_epsg(city: str):
  """根据城市查询epsg代码，用于投影"""
  return epsg_from_lnglat(lng_from_city(city))


def _projection_lnglat(lnglat: (tuple, list), crs_from, crs_to):
  """对经纬度进行投影"""
  from pyproj import Transformer
  if any([is_empty(i) for i in lnglat]):
    return np.nan, np.nan
  transformer = Transformer.from_crs(crs_from, crs_to)
  return transformer.transform(xx=lnglat[1], yy=lnglat[0])


@check_str
def wkb_loads(x: str, hex=True):
  """将文本形式的WKB转换为Shapely几何对象"""
  try:
    return wkb.loads(x, hex=hex)
  except (AttributeError, WKBReadingError) as e:
    warnings.warn(f'{e}, 【{x}】')


@check_shapely
def wkb_dumps(x: BaseGeometry, hex=True, srid=4326):
  """将Shapely几何对象转换为文本形式的WKB"""
  return wkb.dumps(x, hex=hex, srid=srid)


@check_str
def wkt_loads(x: str):
  """将文本形式的WKT转换为Shapely几何对象"""
  try:
    return wkt.loads(x)
  except (AttributeError, WKTReadingError, TypeError) as e:
    warnings.warn(f'{e}, 【{x}】')


@check_shapely
def wkt_dumps(x: BaseGeometry):
  """将Shapely几何对象转换为文本形式的WKT"""
  return wkt.dumps(x)


@check_null()
def geojson_loads(x: (str, dict), warning=True):
  """geojson文本形式转为shapely几个对象"""
  from simplejson.errors import JSONDecodeError
  _shape_e = (AttributeError, GeometryTypeError)
  _x = x
  if isinstance(x, str):
    try:
      return shape(geojson.loads(x))
    except (*_shape_e, JSONDecodeError):
      try:
        return shape(literal_eval(x))
      except (*_shape_e, ValueError, SyntaxError):
        pass
  if isinstance(x, dict):
    try:
      return shape(x)
    except _shape_e:
      pass
  if warning:
    warnings.warn(f'无法转换:"{_x}"')


@check_shapely
def geojson_dumps(x: BaseGeometry):
  """shapely转为geojson文本格式"""
  return geojson.dumps(x)


def is_shapely(x, na=False) -> bool:
  """判断是否为shapely格式"""
  if is_empty(x):
    return na
  if isinstance(x, BaseGeometry):
    return True
  return False


def is_wkb(x, na=False) -> bool:
  """判断是否为wkb格式"""
  if is_empty(x):
    return na
  if not isinstance(x, str) or not is_hex(x):
    return False
  try:
    wkb.loads(x, hex=True)
    return True
  except WKBReadingError:
    return False


def is_wkt(x, na=False) -> bool:
  """判断是否为wkt格式"""
  if is_empty(x):
    return na
  if not isinstance(x, str) or not re.match('^[MPL]', x.lstrip(' (')):
    return False
  try:
    wkt.loads(x)
    return True
  except WKTReadingError:
    return False


def is_geojson(x, na=False) -> bool:
  """判断是否为geojson格式"""
  if is_empty(x):
    return na
  if geojson_loads(x, warning=False):
    return True
  return False


@check_null(default_rv='unknown')
def _infer_geom_format(x):
  """推断geometry格式"""
  if is_shapely(x):
    return 'shapely'
  if is_wkb(x):
    return 'wkb'
  if is_wkt(x):
    return 'wkt'
  if is_geojson(x):
    return 'geojson'
  return 'unknown'


@check_null(default_rv='unknown')
def infer_geom_format(series: (str, list, tuple, pd.Series, BaseGeometry)):
  """推断geometry格式"""
  assert isinstance(series, (str, list, tuple, pd.Series, BaseGeometry))
  if isinstance(series, (list, tuple, pd.Series)):
    for i in series:
      if not_empty(i):
        return _infer_geom_format(i)
  return _infer_geom_format(series)


@check_shapely
def ensure_multi_geom(
    geom: (BaseGeometry, BaseMultipartGeometry)
) -> BaseMultipartGeometry:
  """将Point/LineString/Polygon转为multi分别转为MultiPoint/MultiLineString/MultiPolygon"""
  if isinstance(geom, BaseMultipartGeometry):
    return geom
  if isinstance(geom, LineString):
    return MultiLineString([geom])
  if isinstance(geom, Polygon):
    return MultiPolygon([geom])
  if isinstance(geom, Point):
    return MultiPoint([geom])
  raise TypeError(f'无法识别的类型{type(geom)}')


def multiline2multipolygon(multiline_shapely: MultiLineString, force=False):
  """multiline转为multipolygon，直接首尾相连"""
  coords = []
  for line in multiline_shapely.geoms:
    lngs, lats = line.xy
    for i in range(len(lngs)):
      lng, lat = lngs[i], lats[i]
      point = Point((lng, lat))
      if coords and point == coords[-1]:
        continue
      coords.append(point)
  if len(coords) <= 2:
    return
  if not force and coords[0].distance(coords[-1]) >= 0.000001:
    return
  try:
    return MultiPolygon([Polygon(coords)])
  except Exception as e:
    warnings.warn(f'{e}，{multiline_shapely}')
    return


def get_inner_point(polygon: Polygon, within=True):
  """返回面内的一个点，默认返回中心点，当中心点不在面内则返回面内一个点"""
  if is_empty(polygon):
    return
  point = polygon.centroid
  if not polygon.is_valid:
    polygon = polygon.buffer(0.000001)
  if polygon.contains(point) or not within:
    return point
  return polygon.representative_point()


def is_valid_lnglat(lng, lat):
  """判断经纬度是否有效"""
  return -180 <= lng <= 180 and -90 <= lat <= 90


def auto_loads(x) -> BaseGeometry:
  """自动识别地理格式并转换为shapely格式"""
  geom_format = infer_geom_format(x)
  assert geom_format in GEOM_FORMATS, '未知的地理格式'
  if geom_format == 'shapely':
    return x
  return getattr(sys.modules[__name__], f'{geom_format}_loads')(x)


def dumps2x(x: BaseGeometry, geom_format):
  """转换geometry格式"""
  assert geom_format in GEOM_FORMATS, '未知的地理格式'
  x = auto_loads(x)
  if geom_format == 'shapely':
    return x
  return getattr(sys.modules[__name__], f'{geom_format}_dumps')(x)


def ensure_lnglat(lnglat) -> tuple:
  """确保经纬度是tuple类型，如果是geometry会自动提取经纬度"""
  if is_empty(lnglat):
    return np.nan, np.nan
  if isinstance(lnglat, (list, tuple)):
    assert isinstance_in_list(lnglat, (float, int)), '数据类型错误，经度和纬度都应该为数值型'
    assert is_valid_lnglat(*lnglat), '经纬度超出范围'
    return tuple(lnglat)
  if geom := auto_loads(lnglat):
    return geom.centroid.x, geom.centroid.y
  raise ValueError('未知的地理类型')


def distance(
    p1: (tuple, str),
    p2: (tuple, str),
    city: str = None,
    epsg_from: int = 4326,
    epsg_to: (str, int) = None):
  """
  计算两个点（经度，纬度）之间的距离，单位：米

  Args:
    p1: 点位1，经纬度或geometry
    p2: 点位2，经纬度或geometry
    city: 所在城市，用于投影
    epsg_from: epsg代码
    epsg_to: 投影epsg代码
  """
  p1, p2 = ensure_lnglat(p1), ensure_lnglat(p2)
  if not epsg_to:
    epsg_to = get_epsg(city) if city else epsg_from_lnglat([p1[0], p2[0]])
  p1 = _projection_lnglat(p1, epsg_from, epsg_to)
  p2 = _projection_lnglat(p2, epsg_from, epsg_to)
  return Point(p1).distance(Point(p2))


@check_null(default_rv=[])
def split_multi_geoms(geometry: BaseMultipartGeometry) -> list:
  """
  返回多部件要素中的元素（LineString、Point、Polygon）组成的列表

  Args:
    geometry: geometry类型的数据
  """
  geometry = auto_loads(geometry)
  geometry = ensure_multi_geom(geometry)
  return [g for g in geometry.geoms]


def text2lnglats(text: str, point_sep: str, lnglat_sep: str):
  """
  将一段坐标文本信息整理成经纬度列表

  Args:
    text: 要转换为文本
    point_sep: 点位之间的分隔符
    lnglat_sep: 经纬度之间的分隔符
  """
  if point_sep != ' ' and lnglat_sep != ' ':
    text = text.replace(' ', '')
  assert point_sep in text, f'未找到点位分隔符“{point_sep}”'
  assert lnglat_sep in text, f'未找到经纬度分隔符“{lnglat_sep}”'

  text = text.strip().strip(',').strip(point_sep)
  points = text.split(point_sep)
  res = []
  for p in points:
    lnglat = p.split(lnglat_sep)
    lnglat = [float(i) for i in lnglat if i != '']
    res.append(lnglat)
  return [i for i, _ in groupby(res)]


def deg_to_decimal(x: (str, list, tuple)):
  """
  将度分秒格式的经纬度转为小数

  Args:
    x: 字符串或列表，
      - 字符串格式为 123°5'6.77"（分和秒分别为单双引号）
      - 列表长度为3，分别是度分秒，数值型
  """
  if isinstance(x, str):
    x = x.strip('"')
    d, m_s = x.split('°', 1)
    m, s = m_s.split("'")
  elif isinstance(x, (tuple, list)):
    assert len(x) == 3, '长度必须为3'
    d, m, s = x
  else:
    raise TypeError('类型错误，必须为字符串或列表')
  d, m, s = [float(i) for i in (d, m, s)]
  return d + m / 60 + s / 3600


def text2shapely(
    text: str,
    geometry_type: str,
    point_sep: str = ';',
    lnglat_sep: str = ',',
    ensure_multi: bool = True):
  """
  文本转为shapely

  Args:
    text: 坐标组成的文本
    geometry_type: 要输出的地理类型，可选值为 'polygon'、'line'
    point_sep: 点位键的分隔符，默认为分号 ';'
    lnglat_sep: 经纬度的分隔符，默认为逗号 ','
    ensure_multi: 是否转为multi-geometry，默认为True

  Examples:
    >>> ss = '121.4737,31.2304; 121.4740,31.2304; 121.4740,31.2307'
    >>> text2shapely(ss, geometry_type='line', point_sep=';', lnglat_sep=',')
    MULTILINESTRING ((121.4737 31.2304, 121.474 31.2304, 121.474 31.2307))
  """

  geometry_type = geometry_type.lower()
  assert geometry_type in ('linestring', 'polygon', 'line')

  points = text2lnglats(text, point_sep=point_sep, lnglat_sep=lnglat_sep)

  if geometry_type == 'polygon':
    res = Polygon(points)
  else:
    res = LineString(points)

  if ensure_multi:
    return ensure_multi_geom(res)
  return res


def filter_polygon_from_collection(x):
  """从GeometryCollection中筛选面重新组成geometry"""
  if isinstance(x, (Polygon, MultiPolygon)):
    return x
  if isinstance(x, GeometryCollection):
    return unary_union([
      i for i in x.geoms if isinstance(i, (Polygon, MultiPolygon))
    ])
  raise TypeError(f'不支持的数据类型:{type(x)}')


def ensure_valid_polygon(x):
  if not x.is_valid:
    valid_res = make_valid(x)
    return filter_polygon_from_collection(valid_res)
  return x



import logging
import warnings

import pandas as pd
from shapely.geometry.base import BaseGeometry


def ensure_list(val):
  """将标量值和Collection类型都统一转换为LIST类型"""
  if val is None:
    return []
  if isinstance(val, list):
    return val
  if isinstance(val, (set, tuple)):
    return list(val)
  if isinstance(val, pd.Series):
    return val.tolist()
  return [val]


def ensure_ext(ext: str):
  """
  规范文件扩展名，保持扩展名为点(.)开头

  Examples:
    >>> ensure_ext('csv')
    '.csv'
  """
  if isinstance(ext, str) and ext != '' and not ext.startswith('.'):
    return '.' + ext
  return ext


def is_empty(x) -> bool:
  """
  判断变量是否为空值

  以下值认为是空白：
    - 空白列表、字典, 如：`[]`, `{}`
    - 空白Dataframe、Series, 如：`pd.DataFrame()`
    - 空白shapely格式的geometry，如：`Point(np.nan, np.nan)`
  """
  if isinstance(x, (list, dict, tuple)):
    return False if x else True
  if isinstance(x, (pd.DataFrame, pd.Series)):
    return x.empty
  if isinstance(x, BaseGeometry):
    return x.is_empty
  return pd.isna(x)


def not_empty(x) -> bool:
  """判断是否非空，对`is_empty`取反"""
  return not is_empty(x)


def second_to_dhms(second) -> tuple:
  """将秒转为天、时、分、秒"""
  d = int(second // 86400)
  second = second % 86400
  h = int(second // 3600)
  second = second % 3600
  m = int(second // 60)
  second = second % 60
  if d + h + m == 0:
    second = round(second, 2)
  else:
    second = int(second)
  return d, h, m, second


def second_to_desc(second) -> str:
  """
  将秒转为时间描述

  Examples:
    >>> second_to_desc(123)
    '2m 3s'
    >>> second_to_desc(1234)
    '20m 40s'
    >>> second_to_desc(123456)
    '1d 10h 17m 36s'
  """
  d, h, m, second = second_to_dhms(second)
  if d + h + m == 0:
    return f'{second:.2f}s'
  if d + h == 0:
    return f'{m}m {second:.0f}s'
  if d == 0:
    return f'{h}h {m}m {second:.0f}s'
  return f'{d}d {h}h {m}m {second:.0f}s'


def warn_(msg, if_or_not=True, mode='warning'):
  if if_or_not:
    if mode == 'warning':
      warnings.warn(msg)
    if mode == 'logging':
      logging.warning(msg)


def agg_parser(agg: dict) -> list:
  """将字典形式的agg聚合参数转换为[原字段名，计算函数，储存字段名]列表形式"""
  res = []
  for c, functions in agg.items():
    functions = ensure_list(functions)
    for func in functions:
      res.append([c, func, f'{c}_{func}'])
  return res



from functools import lru_cache

from .etl.extract import rdf

BASE_PATH = '/Users/ricco/common_data'


class Ricco:
  """通用数据"""

  @lru_cache
  def bd_region(self):
    return rdf(f'{BASE_PATH}/bd_region')

  def region(self, city='上海'):
    return rdf(f'{BASE_PATH}/行政边界/{city}边界_区县最新.csv')

  def street(self, city='上海'):
    return rdf(f'{BASE_PATH}/行政边界/{city}边界_街镇最新.csv')


Rc = Ricco()



def fix_polygon_topology(
    gdf, geo_col='geometry', fill_intersects=False, keep_contains=False  # noqa
):
  raise Exception(
      '`fix_polygon_topology`已经停用，'
      '请使用`ricco.geometry.topology.fix_topology`代替。'
  )



import sys
import warnings
from typing import List
from typing import Union

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Polygon
from shapely.geometry.base import BaseGeometry

from ..base import agg_parser
from ..base import ensure_list
from ..base import not_empty
from ..base import warn_
from ..etl.transformer import dict2df
from ..etl.transformer import split_list_to_row
from ..util.assertion import assert_not_null
from ..util.assertion import assert_series_unique
from ..util.decorator import process_multi
from ..util.decorator import progress
from ..util.decorator import timer
from ..util.kdtree import kdtree_nearest
from ..util.util import first_notnull_value
from .util import GEOM_FORMATS
from .util import auto_loads
from .util import epsg_from_lnglat
from .util import geojson_dumps
from .util import geojson_loads
from .util import get_epsg
from .util import infer_geom_format
from .util import split_multi_geoms
from .util import wkb_dumps
from .util import wkb_loads
from .util import wkt_dumps
from .util import wkt_loads


def projection(
    df: gpd.GeoDataFrame,
    epsg: int = None,
    city: str = None,
    crs=None,
    c_geometry='geometry'
) -> gpd.GeoDataFrame:
  """
  投影变换

  Args:
    df: 输入的GeomDataFrame格式的数据
    epsg: epsg code, 第一优先级
    city: 城市名称，未传入epsg的情况下将通过城市名称获取epsg，若二者都为空则根据经纬度获取
    crs: 投影坐标系，第二优先级
    c_geometry: geometry的列名
  """
  df = auto2shapely(df, geometry=c_geometry)
  if not epsg and not crs:
    if city:
      epsg = get_epsg(city)
    else:
      df_temp = df.bounds
      lng = df_temp[['minx', 'maxx']].mean(axis=1).median()
      epsg = epsg_from_lnglat(lng)
      print(f'从数据集中自动获取的epsg code为：{epsg}')
  return df.to_crs(epsg=epsg, crs=crs)


@timer()
def projection_lnglat(
    df: pd.DataFrame,
    epsg=None,
    city=None,
    crs=None) -> pd.DataFrame:
  """
  直接对经纬度进行投影变换

  Args:
    df: 输入的GeomDataFrame格式的数据
    epsg: epsg code, 第一优先级
    city: 城市名称，未传入epsg的情况下将通过城市名称获取epsg，若二者都为空则根据经纬度获取
    crs: 投影坐标系，第二优先级
  """
  df = df.copy()
  df_temp = lnglat2shapely(df[['lng', 'lat']])
  df_temp = projection(df_temp, epsg=epsg, city=city, crs=crs)
  df_temp = shapely2lnglat(df_temp[['geometry']])[['lng', 'lat']]
  df.update(df_temp)
  if 'geometry' in df:
    warnings.warn('仅对lng, lat列进行投影变换，未对geometry进行投影变换')
  return df


@progress
def wkb2shapely(df,
                geometry='geometry',
                epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """将wkb格式的geometry列转换为shapely格式"""
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(wkb_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


@progress
def shapely2wkb(df, geometry='geometry'):
  """将shapely格式的geometry列转换为wkb格式"""
  df = pd.DataFrame(df).copy()
  df[geometry] = df[geometry].progress_apply(wkb_dumps)
  return df


@progress
def wkt2shapely(df,
                geometry='geometry',
                epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """将wkt格式的geometry列转换为shapely格式"""
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(wkt_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


@progress
def shapely2wkt(df, geometry='geometry'):
  """将shapely格式的geometry列转换为wkt格式"""
  df = pd.DataFrame(df).copy()
  df[geometry] = df[geometry].progress_apply(wkt_dumps)
  return df


@progress
def geojson2shapely(df,
                    geometry='geometry',
                    epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """将geojson格式的geometry列转换为shapely格式"""
  df = df.copy()
  df[geometry] = df[geometry].progress_apply(geojson_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


@progress
def shapely2geojson(df, geometry='geometry'):
  """将shapely格式的geometry列转换为geojson格式"""
  df = pd.DataFrame(df).copy()
  df[geometry] = df[geometry].progress_apply(geojson_dumps)
  return df


@progress
def lnglat2shapely(df,
                   lng='lng',
                   lat='lat',
                   geometry='geometry',
                   delete=True,
                   epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """将经纬度坐标转换为shapely格式的geometry"""
  df = gpd.GeoDataFrame(
      df,
      geometry=gpd.points_from_xy(df[lng], df[lat]),
      crs=epsg_code
  )
  df.rename(columns={'geometry': geometry}, inplace=True)
  if delete:
    del df[lng], df[lat]
  return gpd.GeoDataFrame(df, geometry=geometry)


def shapely2lnglat(df,
                   geometry='geometry',
                   lng='lng',
                   lat='lat',
                   within=False,
                   delete=False):
  """
  shapely格式提取面内点转为经纬度。

  Args:
    df: 要转换的DataFrame
    geometry: 输入的geometry列名
    lng: 输出的经度列名
    lat: 输出的纬度列名
    within: 范围的点是否再面内
      - False(default): 直接返回中心点；
      - True: 返回面内的一个点
    delete: 是否删除geometry
  """

  df = df.copy()
  if within:
    df[lng] = df.representative_point().x
    df[lat] = df.representative_point().y
  else:
    df[lng] = df.centroid.x
    df[lat] = df.centroid.y
  if delete:
    del df[geometry]
  return df


def wkb2lnglat(df, geometry='geometry', delete=False, within=False):
  """geometry转经纬度，求中心点经纬度"""
  df = wkb2shapely(df, geometry=geometry)
  df = shapely2lnglat(df, geometry=geometry, within=within, delete=delete)
  if not delete:
    df = shapely2wkb(df, geometry=geometry)
  return df


def lnglat2wkb(df,
               lng='lng',
               lat='lat',
               geometry='geometry',
               delete=False, epsg_code=4326):
  """经纬度转wkb格式的geometry"""
  df = lnglat2shapely(
      df, lng, lat, geometry=geometry, delete=delete, epsg_code=epsg_code
  )
  df = shapely2wkb(df, geometry=geometry)
  return df


def wkt2lnglat(df, geometry='geometry', delete=False, within=False):
  """geometry转经纬度，求中心点经纬度"""
  df = wkt2shapely(df, geometry=geometry)
  df = shapely2lnglat(df, geometry=geometry, within=within, delete=delete)
  if not delete:
    df = shapely2wkt(df, geometry=geometry)
  return df


def lnglat2wkt(df,
               lng='lng',
               lat='lat',
               geometry='geometry',
               delete=False, epsg_code=4326):
  """经纬度转wkb格式的geometry"""
  df = lnglat2shapely(
      df, lng, lat, geometry=geometry, delete=delete, epsg_code=epsg_code
  )
  df = shapely2wkt(df, geometry=geometry)
  return df


def wkb2wkt(df, geometry='geometry', epsg_code: int = 4326):
  """wkb转wkt"""
  df = wkb2shapely(df, geometry=geometry, epsg_code=epsg_code)
  return shapely2wkt(df, geometry=geometry)


def wkt2wkb(df, geometry='geometry', epsg_code: int = 4326):
  """wkb转wkt"""
  df = wkt2shapely(df, geometry=geometry, epsg_code=epsg_code)
  return shapely2wkb(df, geometry=geometry)


def shapely2central_shapely(df, geometry='geometry', within=False):
  """获取中心点shapely格式"""
  df = shapely2lnglat(df, geometry=geometry, within=within)
  return lnglat2shapely(df, geometry=geometry, delete=True)


def auto2shapely(df, geometry='geometry') -> gpd.GeoDataFrame:
  """自动识别地理格式并转换为shapely格式"""
  if geometry in df:
    geom_format = infer_geom_format(df[geometry])
    assert geom_format in GEOM_FORMATS, '未知的地理格式'
    if geom_format == 'shapely':
      return gpd.GeoDataFrame(df, geometry=geometry)
    return getattr(sys.modules[__name__],
                   f'{geom_format}2shapely')(df, geometry=geometry)
  elif 'lng' in df and 'lat' in df:
    warn_('未找到geometry列，尝试将lng/lat列转换为shapely格式')
    return lnglat2shapely(df, delete=False)
  else:
    raise KeyError(f'未找到{geometry}或lng/lat列')


def shapely2x(df: (gpd.GeoDataFrame, pd.DataFrame),
              geometry_format: str,
              geometry='geometry'):
  """
  将shapely转为指定的格式

  Args:
    df: 要转换的GeoDataFrame
    geometry_format: 支持wkb,wkt,shapely,geojson
    geometry: geometry列的列名，默认“geometry”
  """
  assert geometry_format in GEOM_FORMATS, '未知的地理格式'
  if geometry_format == 'shapely':
    return gpd.GeoDataFrame(df, geometry=geometry)
  return getattr(sys.modules[__name__],
                 f'shapely2{geometry_format}')(df, geometry=geometry)


def auto2x(df, geometry_format: str, geometry='geometry'):
  """
  将geometry转为指定格式

  Args:
    df: 要转换的Dataframe
    geometry_format: 要转换为的geometry类型，支持shapely,wkb,wkt,geojson
    geometry: geometry列的列名，默认为“geometry”
  """
  assert geometry_format in GEOM_FORMATS, '未知的地理格式'
  if infer_geom_format(df[geometry]) == geometry_format:
    return df
  df = auto2shapely(df, geometry=geometry)
  return shapely2x(df, geometry_format=geometry_format, geometry=geometry)


def distance_min(df: gpd.GeoDataFrame, geometry: BaseGeometry) -> float:
  """
  计算数据集df中元素到单个geometry的最短距离

  Args:
    df: 数据集，GeoDataFrame格式
    geometry: 单个geometry，shapely格式
  """
  df = auto2shapely(df)
  geometry = auto_loads(geometry)
  return df.distance(geometry).min()


@process_multi
def distance_gdf(df: gpd.GeoDataFrame, df_target: gpd.GeoDataFrame,
                 c_dst: str = '最小距离',
                 left_geometry='geometry'):
  """计算一个数据集中的每个元素到另一个数据集之间的最短距离"""
  assert df.index.is_unique, 'df 索引列必须唯一'
  df_target = auto2shapely(df_target)
  df_dis = auto2shapely(df[[left_geometry]], geometry=left_geometry)
  df_dis[c_dst] = df_dis[left_geometry].parallel_apply(
      lambda p: distance_min(df_target, p) if not_empty(p) else np.nan
  )
  return df.join(df_dis[[c_dst]])


def split_grids(df: gpd.GeoDataFrame, step: int, geometry_format='wkb'):
  """
  根据所给边界划分固定边长的栅格

  Args:
    df: 边界文件，GeoDataFrame格式
    step: 栅格边长，单位：米
    geometry_format: 栅格格式，支持wkb,wkt,shapely,geojson
  """

  def get_xxyy(_df):
    _df = _df.bounds
    return [
      _df['maxx'][0] - _df['minx'][0], _df['maxy'][0] - _df['miny'][0],
      _df['minx'][0], _df['miny'][0], _df['maxx'][0], _df['maxy'][0]
    ]

  # 融合
  df = auto2shapely(df)
  df = df[['geometry']].dissolve().reset_index(drop=True)
  df.crs = 'epsg:4326'
  # 投影，获取实际的距离
  df_p = projection(df)
  # 分别获取投影前和投影后的点位及边长信息
  xy_o, xy_p = get_xxyy(df), get_xxyy(df_p)
  # 获取栅格的起止经纬度，步长等信息
  xr, yr = xy_o[0] / xy_p[0], xy_o[1] / xy_p[1]
  x_len, y_len = xr * step, yr * step
  x_num, y_num = int(xy_o[0] // x_len + 1), int(xy_o[1] // y_len + 1)
  warn_(f'lng分段数：{x_num}，lat分段数：{y_num}', mode='logging')
  x_start, y_start = xy_o[2], xy_o[3]
  x_end, y_end = x_start + x_num * x_len, y_start + y_num * y_len
  # 构建经纬度列表
  lng_array = np.linspace(x_start, x_end, num=x_num)
  lat_array = np.linspace(y_start, y_end, num=y_num)
  lng_list, lat_list = list(lng_array), list(lat_array)
  # 构建栅格索引
  res = {}
  for i in range(len(lng_list) - 1):
    for j in range(len(lat_list) - 1):
      res[f'{i + 1}-{j + 1}'] = Polygon.from_bounds(
          lng_list[i], lat_list[j], lng_list[i + 1], lat_list[j + 1]
      )
  # 通过边界裁剪栅格
  df_res = gpd.GeoDataFrame(
      dict2df(res, 'grid_id', 'geometry'), crs="epsg:4326")

  df_res = gpd.sjoin(
      df_res, df, how='inner',
      predicate='intersects').drop(
      'index_right', axis=1
  )
  return auto2x(df_res, geometry_format=geometry_format).reset_index(drop=True)


def ensure_geometry(df,
                    ensure_point=False,
                    warning=False,
                    geometry='geometry',
                    lng='lng',
                    lat='lat'):
  """转换并仅保留geometry列"""
  if geometry in df:
    if lng in df and lat in df:
      if not df[df[lng].notna() & df[geometry].isna()].empty:
        warn_(f'存在“{geometry}”为空但经纬度不为空的行', warning)
    df = auto2shapely(df[[geometry]], geometry=geometry)[[geometry]]
    if ensure_point:
      geom = first_notnull_value(df[geometry])
      if geom.geom_type not in ['point', 'Point']:
        warn_('非点数据，提取面内点', warning)
        df = shapely2central_shapely(df, within=True, geometry=geometry)
    return df
  if lng in df and lat in df:
    return lnglat2shapely(
        df[[lng, lat]],
        lng=lng, lat=lat,
        geometry=geometry,
        delete=False,
    )[[geometry]]

  raise KeyError(f'文件中必须有{lng},{lat}列或{geometry}列')


def ensure_lnglat(df, lng='lng', lat='lat', geometry='geometry'):
  """自动提取点或转换为点的经纬度"""
  if lng in df and lat in df:
    return df
  if geometry in df:
    df_temp = auto2shapely(df[[geometry]])
    df_temp = shapely2lnglat(df_temp, within=True)
    return df.join(df_temp[[lng, lat]])
  raise AssertionError('无可转为经纬度的列')


def mark_tags_v2(
    point_df: pd.DataFrame,
    polygon_df: pd.DataFrame,
    col_list: (list, str) = None,
    *,
    predicate='intersects',
    drop_geometry=False,
    geometry_format='wkb',
    warning=True,
    point_lng='lng',
    point_lat='lat',
    point_geometry='geometry',
    polygon_geometry='geometry',
    warning_message=None,
    ensure_point=True
):
  """
  使用面数据通过空间关联（sjoin）给数据打标签

  Args:
    point_df: 要打标签的数据，一般为点数据，如为面数据，则会自动提取面内点计算
    polygon_df: 标签所在的数据集，一般为面数据
    col_list: 面数据中要关联到结果中的列，若为空则全部关联
    predicate: 关联方法，默认 'intersects'
    drop_geometry: 结果是否删除point_df中的geometry，默认不删除
    geometry_format: 输出的geometry格式，支持wkb,wkt,shapely,geojson，默认wkb
    warning: 是否输出警告信息
    point_lng: 指定点数据的经度列名
    point_lat: 指定点数据的经度列名
    point_geometry: 指定点数据的geometry列名
    polygon_geometry: 指定面数据的geometry列名
    warning_message: 是否输出警告信息，已弃用
    ensure_point: point_df是否强制转换为点数据
  """
  if warning_message is not None:
    warn_('参数"warning_message"已弃用，请使用参数"warning"')

  if point_df.empty or point_df.empty:
    warn_('存在空的数据集，请检查', warning)
    return point_df

  point_df = point_df.copy()
  assert point_df.index.is_unique, 'point_df索引列必须唯一'
  if not col_list:
    col_list = polygon_df.columns.to_list()
  else:
    col_list = ensure_list(col_list)
    polygon_df = polygon_df[[*col_list, polygon_geometry]]

  if cols_mapping := {
    c: f'{c}_origin' for c in col_list
    if c in point_df and c not in [point_lng, point_lat, point_geometry]
  }:
    warn_(f'同名字段重命名：{cols_mapping}', warning)
    point_df.rename(columns=cols_mapping, inplace=True)

  # 转换为shapely格式
  _df = ensure_geometry(point_df, ensure_point, warning,
                        lng=point_lng, lat=point_lat, geometry=point_geometry)
  polygon_df = auto2shapely(polygon_df, geometry=polygon_geometry)
  # 空间关联
  _df = _df.sjoin(
      polygon_df, how='left', predicate=predicate,
  ).drop(['index_right'], axis=1)
  del _df[point_geometry]
  # 统一geometry输出格式、删除geometry、避免多次转换
  if point_geometry in point_df:
    if drop_geometry:
      del point_df[point_geometry]
    else:
      point_df = auto2x(point_df, geometry_format, geometry=point_geometry)
  # 将空间关联后的数据关联到原来的Dataframe上
  return point_df.join(_df, how='left')


def nearest_neighbor(
    df: pd.DataFrame,
    df_target: pd.DataFrame,
    c_dst: str = 'min_distance',
    epsg: int = None) -> pd.DataFrame:
  """
  近邻分析，计算一个数据集中的元素到另一个数据集中全部元素的最短距离（单位：米）

  Args:
    df:
    df_target:
    c_dst: 输出最短距离的列名，默认为“min_distance”
    epsg: 对于跨时区或不在同一个城市的可以指定epsg code，默认会根据经度中位数获取
  """
  # 将两个数据集都转为shapely格式
  assert df.index.is_unique, 'df索引列必须唯一'
  df_left = ensure_geometry(df)
  df_target = ensure_geometry(df_target)
  # 投影
  df_left = projection(df_left, epsg=epsg)
  df_target = projection(df_target, epsg=epsg)
  # 计算最短距离
  df_left = distance_gdf(df_left, df_target, c_dst=c_dst)
  # 将距离合并到原来的数据集上
  return df.join(df_left[[c_dst]], how='left')


@progress
def nearest_kdtree(
    df: pd.DataFrame,
    df_poi: pd.DataFrame,
    /, *,
    c_count: str = 'count',
    c_min_distance: str = 'min_distance',
    agg: dict = None,
    limit: int = None,
    r: (int, float) = None,
    keep_origin: bool = False,
    leaf_size: int = 2,
    epsg: int = None,
    city=None,
    crs=None,
):
  """
  KDTree近邻分析，计算一个数据集中的元素到另一个数据集中全部元素的最短距离（单位：米）,
  同时可进行其他运算

  Args:
    df: 基础数据集，统计该数据及周边的其他数据集的信息
    df_poi: 被统计的数据集
    c_count: 计数列字段名，默认“count”
    c_min_distance: 最短距离字段名，默认“min_distance”
    agg: 计算 df_poi 中的其他字段，格式如: {'面积': ['sum', 'mean']}，即计算面积的和、均值
    limit: 限制符合条件的 df_poi 中的个数，由近及远
    r: 限制查询半径
    keep_origin: 是否保留匹配后原始的索引信息
    leaf_size: KDTree 叶子节点大小
    epsg: 投影代码，用于投影，epsg/city/crs指定多个时以epsg为准
    city: 城市，用于获取城市中心点，epsg/city/crs指定多个时以epsg为准
    crs: 数据集的 crs，epsg/city/crs指定多个时以epsg为准
  """
  assert df.index.is_unique and df_poi.index.is_unique, '数据集索引必须唯一'
  # 确保数据中有经纬度
  df_temp = ensure_lnglat(df)
  df_poi = ensure_lnglat(df_poi)

  df_temp = df_temp[df_temp.lng.notna() & df_temp.lat.notna()]
  df_poi = df_poi[df_poi.lng.notna() & df_poi.lat.notna()]
  assert not df_temp.empty and not df_poi.empty, '筛选后数据集为空，请检查经纬度列'
  # 投影
  df_temp = projection_lnglat(
      df_temp, epsg=epsg, city=city, crs=crs
  )[['lng', 'lat']]
  df_poi = projection_lnglat(df_poi, epsg=epsg, city=city, crs=crs)
  # 计算（范围内）的一个点或多个点
  data_tree = df_poi[['lng', 'lat']].values
  data_query = df_temp[['lng', 'lat']].values
  df_temp['ind_list'], df_temp['dist_list'] = kdtree_nearest(
      data_tree, data_query, limit=limit, r=r, leaf_size=leaf_size,
  )
  # 统计数量
  if c_count:
    assert c_count not in df, f'"{c_count}"已存在，请更换列名'
    df_temp[c_count] = df_temp['ind_list'].progress_apply(lambda x: len(x))
  # 计算最短距离
  if c_min_distance:
    assert c_min_distance not in df, f'"{c_min_distance}"已存在，请更换列名'
    df_temp[c_min_distance] = df_temp['dist_list'].progress_apply(
        lambda x: min(x) if len(x) > 0 else None
    )
  # 计算其他字段
  if agg:
    for c_src, func, c_dst in agg_parser(agg):
      df_temp[c_dst] = df_temp['ind_list'].progress_apply(
          lambda x: getattr(df_poi.loc[x][c_src], func)()
      )
  # 保留关联内容
  if not keep_origin:
    del df_temp['ind_list'], df_temp['dist_list']
  del df_temp['lng'], df_temp['lat']
  return df.join(df_temp)


def get_neighbors(
    df: (pd.DataFrame, gpd.GeoDataFrame),
    key_col: str,
    res_type='dict',
) -> (dict, pd.DataFrame):
  """
  获取数据集中每个点的邻居面(相邻标准：至少有一个点相同且不重叠)

  Args:
    df:
    key_col: 关键列，必须唯一
    res_type: 返回类型，可返回dict或Dataframe
  """
  if not key_col:
    key_col = 'index'
    df = df.reset_index()
  assert_series_unique(df, key_col)
  assert_not_null(df, key_col)
  if isinstance(res_type, str):
    res_type = res_type.lower()
  res = {}
  df = auto2shapely(df.copy())
  for key in df[key_col].unique():
    neighbors = []
    v = df[df[key_col] == key].geometry.squeeze()
    for _, row in df.iterrows():
      if v.touches(row['geometry']):
        neighbors.append(row[key_col])
      if neighbors:
        res[key] = neighbors
  if res_type in (dict, 'dict'):
    return res
  if res_type in (pd.DataFrame, 'df', 'dataframe'):
    df = dict2df(res)
    return df.explode('value')
  raise ValueError(f'错误的res_type:{res_type}')


def get_area(
    df: pd.DataFrame,
    c_dst='area',
    epsg: int = None,
    decimals=2) -> pd.DataFrame:
  """
  计算面积（单位：平方米）

  Args:
    df: 要计算的面数据
    c_dst: 输出面积的列名，默认为“area”
    epsg: 对于跨时区或不在同一个城市的可以指定epsg code，默认会根据经度中位数获取
    decimals: 要保留的小数位数
  """
  # 将数据集转为shapely格式
  assert df.index.is_unique, 'df索引列必须唯一'
  assert c_dst not in df, f'"{c_dst}"列已存在，请指定不同的c_dst'
  df_left = ensure_geometry(df)
  # 投影
  df_left = projection(df_left, epsg=epsg)
  # 计算面积
  df_left[c_dst] = df_left.area.round(decimals)
  # 将面积合并到原来的数据集上
  return df.join(df_left[[c_dst]], how='left')


def buffer(df: pd.DataFrame,
           radius: Union[int, float],
           city: str = None,
           geo_type: str = 'point',
           geometry: str = 'geometry',
           buffer_geometry: str = 'buffer_geometry',
           geo_format='wkb') -> pd.DataFrame:
  """
  获得一定半径的缓冲区

  Args:
    df: pd.DataFrame, 包含地理信息的DataFrame
    radius: numeric, 缓冲区半径（单位米）
    city: str, 可选, 投影城市，可提高数据精度
    geo_type: str, 地理数据类型，可选point, line或polygon(包括multipolygon)，默认point
    geometry: str, geometry字段名，默认"geometry"
    buffer_geometry: 输出的缓冲区geometry字段名，默认"buffer_geometry"
    geo_format: str, 输出的缓冲区geometry格式，支持wkb,wkt,shapely,geojson，默认wkb
  Returns:
    包含缓冲区geometry的DataFrame
  """
  df = df.copy()
  assert df.index.is_unique, 'df索引列必须唯一'
  if geo_type == 'point':
    df_buffer = ensure_geometry(df, True)
  elif geo_type in ['line', 'polygon']:
    df_buffer = auto2shapely(df, geometry=geometry)[[geometry]]
  else:
    raise ValueError('geo_type必须为point，line或polygon')
  crs = df_buffer.crs
  df_buffer = projection(df_buffer, city=city)
  df_buffer[buffer_geometry] = df_buffer.buffer(radius)
  df_buffer = gpd.GeoDataFrame(
      df_buffer[[buffer_geometry]], geometry=buffer_geometry
  )
  df_buffer = projection(df_buffer, crs=crs, c_geometry=buffer_geometry)
  df = df.join(df_buffer, how='left')
  return shapely2x(df, geo_format, geometry=buffer_geometry)


def spatial_agg(point_df: pd.DataFrame,
                polygon_df: pd.DataFrame,
                by: Union[str, List[str]],
                agg: dict,
                polygon_geometry: str = 'geometry') -> pd.DataFrame:
  """
  对面数据覆盖范围内的点数据进行空间统计

  Args:
    point_df: pd.DataFrame, 点数据dataframe;
    polygon_df: pd.DataFrame, 面数据dataframe;
    by: Union[str, List[str]], 空间统计单位字段；
    agg: dict, 空间统计操作。格式为{'被统计字段名': '操作名', ...}的字典。如{'poi':'sum'};
    polygon_geometry: str, 面数据geometry字段名，默认"geometry";
  Returns:
    pd.DataFrame, 包含空间统计单位字段和被统计字段和面数据geometry的DataFrame
  """

  polygon_df = polygon_df[[by, polygon_geometry]]
  point_df = mark_tags_v2(
      point_df=point_df,
      polygon_df=polygon_df,
      col_list=by,
      drop_geometry=True,
      polygon_geometry=polygon_geometry)
  df_grouped = point_df.groupby(by=by, as_index=False).agg(agg)
  return df_grouped


def split_multi_to_rows(df, geometry='geometry', geometry_format=None):
  """将多部件要素拆解为多行的单部件要素"""

  lo = df.shape[0]
  if not geometry_format:
    geometry_format = infer_geom_format(df[geometry])
  df = auto2shapely(df)
  df[geometry] = df[geometry].apply(split_multi_geoms)
  df = split_list_to_row(df, geometry)
  df = gpd.GeoDataFrame(df, geometry=geometry)
  print(f'Rows: {lo} -> {df.shape[0]}')
  return auto2x(df, geometry_format, geometry)



import random
import datetime

import numpy as np
import pandas as pd


def random_name():
  """随机生成中文名字，仅生成2或3字名字"""
  from ..resource.names import FirstName
  from ..resource.names import LastName
  c = [random.choice(FirstName)]
  for i in range(random.randint(1, 2)):
    c.append(random.choice(LastName))
  return ''.join(c)


def random_date():
  """获取随机日期"""
  now = datetime.datetime.now()
  tsp = int(datetime.datetime.timestamp(now))
  tsp = random.randrange(-946800000, tsp)
  return datetime.datetime.fromtimestamp(tsp).strftime("%Y%m%d")


def random_room_number(unit=False):
  """生成随机的房间号"""
  floor = random.randint(1, 17)
  room = random.randint(1, 4)
  room = str(room).zfill(2)
  if unit:
    _u = random.randint(1, 6)
    _unit = f'{_u}单元'
    return f'{_unit}{floor}{room}'
  return f'{floor}{room}'


def random_by_prob(mapping: dict):
  """根据概率生成随机值"""
  return np.random.choice(
      list(mapping.keys()),
      p=np.array(list(mapping.values())).ravel()
  )


def ramdom_lnglat(
    df: pd.DataFrame,
    lng_range: (tuple, list) = (72, 138),
    lat_range: (tuple, list) = (0, 56),
):
  """
  Args:
    df: 输入的dataframe
    lng_range: 城市的经度范围
    lat_range: 城市的纬度范围
  """
  length = df.shape[0]
  df['lng'] = np.random.uniform(*lng_range, length)
  df['lat'] = np.random.uniform(*lat_range, length)
  return df



import time
import warnings
from functools import wraps

from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from ..base import is_empty
from ..base import second_to_desc


def progress(func):
  """tqdm进度条（progress_apply）"""

  @wraps(func)
  def wrapper(*args, **kwargs):
    tqdm.pandas(desc=func.__name__)
    __rv = func(*args, **kwargs)
    tqdm.pandas()
    return __rv

  return wrapper


def process_multi(func):
  """多线程处理apply任务（parallel_apply）"""

  @run_once
  def init_pandarallel():
    from pandarallel import pandarallel
    pandarallel.initialize()

  @wraps(func)
  def wrapper(*args, **kwargs):
    init_pandarallel()
    return func(*args, **kwargs)

  return wrapper


def print_doc(func):
  """打印docstring"""

  @wraps(func)
  def wrapper(*args, **kwargs):
    print(func.__doc__)
    return func(*args, **kwargs)

  return wrapper


def timer(desc=None):
  """函数运行时间统计"""

  def _timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      start_time = time.time()
      result = func(*args, **kwargs)
      end_time = time.time()
      duration = end_time - start_time
      time_desc = second_to_desc(duration)
      title = desc or func.__name__
      print(f'Costs：{duration:.2f} s ({time_desc})，function: {title} ')
      return result

    return wrapper

  return _timer


def check_null(default_rv=None):
  """检查第一个参数是否非空，若为空则直接返回空值"""

  def _check_null(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      if len(args) > 0:
        if is_empty(args[0]):
          return default_rv
      elif len(kwargs) > 0:
        if is_empty(list(kwargs.values())[0]):
          return default_rv
      else:
        raise ValueError('No args or kwargs')
      return func(*args, **kwargs)

    return wrapper

  return _check_null


def check_str(func):
  """检查第一个参数是否是字符串，若非字符串则警告并返回空值"""

  @wraps(func)
  def wrapper(*args, **kwargs):
    if is_empty(args[0]):
      return
    if not isinstance(args[0], str):
      warnings.warn(f'TypeError:【{args[0]}】')
      return
    return func(*args, **kwargs)

  return wrapper


def check_shapely(func):
  """检查第一个参数是否是shapely格式，若非shapely格式则警告并返回空值"""

  @wraps(func)
  def wrapper(*args, **kwargs):
    if is_empty(args[0]):
      return
    if not isinstance(args[0], BaseGeometry):
      warnings.warn(f'输入值不是shapely格式:【{args[0]}】')
      return
    return func(*args, **kwargs)

  return wrapper


def run_once(func):
  """在一次执行中只运行一次，注：除第一次运行外，后续的不会执行也无返回值"""

  @wraps(func)
  def wrapper(*args, **kwargs):
    if not wrapper.has_run:
      wrapper.has_run = True
      return func(*args, **kwargs)

  wrapper.has_run = False
  return wrapper


def as_staticmethod(cls):
  """将函数添加到类中并作为staticmethod"""

  @wraps(cls)
  def decorator(func):
    setattr(cls, func.__name__, staticmethod(func))
    return func

  return decorator
